# capture_picoscope
# requires pynqserver to run on Pynq
# requires FOBOS Shield
# uses Picoscope
import os
import sys
import shutil
import json
import numpy as np
import matplotlib.pyplot as plt
import time
import math
from datetime import datetime

import cpa_bike as cpa
import traceset as traceset
import utils as powermodelUtils
import bike_powermodel as powermodel


def bike_attack(sec_lvl, projDir, PLAIN_TEXT, TRACES_FILE, results_dir, logfile):

    with open("mem.log", "w") as f:
        pass

    PLAIN_TEXT   = os.path.join(projDir, PLAIN_TEXT)
    TRACES_FILE  = os.path.join(projDir, TRACES_FILE)

    if not os.path.exists(results_dir):
        os.makedirs(results_dir)

    flog = open(results_dir+"/"+logfile, "w+")

    print(PLAIN_TEXT,"\n",file=flog)
    print(TRACES_FILE ,"\n",file=flog)

    # Sampling conifiguration
    sampling_freq     = 1000
    target_clk        = 5

    if (sec_lvl == 1):
        ### Parameters Level 1
        r = 12323
        w = 71

        trigger_sim_ns     = 44065
        first_write_sim_ns = 49645
        first_read_sim_ns  = 53545
        second_read_sim_ns = 57445

    elif (sec_lvl == 3):
        ### Parameters Level 3
        r = 24659
        w = 206//2

        trigger_sim_ns     = 79985
        first_write_sim_ns = 90045
        first_read_sim_ns  = 97805
        second_read_sim_ns = 105555

    elif (sec_lvl == 5):
        #### Parameters Level 5
        r = 40973
        w = 274//2

        trigger_sim_ns     = 126225
        first_write_sim_ns = 142075
        first_read_sim_ns  = 154925
        second_read_sim_ns = 167775
    else:
        exit()

    ## Init variables
    known_keys        = []
    known_partials    = []
    mtd_intersections = []

    target_write_cc     = ((first_write_sim_ns - trigger_sim_ns)//10)
    target_write_sample = (target_write_cc)*(sampling_freq//target_clk)

    target_read_cc     = ((first_read_sim_ns - trigger_sim_ns)//10)
    target_read_sample = (target_read_cc)*(sampling_freq//target_clk)

    round_len_cc     = ((second_read_sim_ns-first_read_sim_ns)//10)
    round_len_sample = round_len_cc*(sampling_freq//target_clk)



    ################################################
    ## RUN
    ################################################
    abs_start = time.time()

    c0_dummy, h0_expected = powermodel.loadInput(PLAIN_TEXT,1,r)
    h0_expected = h0_expected[0]
    print(h0_expected,flush=True)
    print(h0_expected,file=flog,flush=True)

    ##############################################
    # Attack first write to fully recover h0[0]
    ##############################################
    # Attack configuration
    traces            = 40
    splice_depth      = math.ceil(r/32)-1

    # ---------
    # Power Model
    # ---------
    hypotheticalPower = [powermodel.getHypotheticalPower(PLAIN_TEXT, traces, splice_depth, 0, [], [], r)]

    # ---------
    # Get Traces
    # ---------
    ### Crop to target cycle
    print("Reading power traces...", "\n",file=flog,flush=True)
    start_sample  = target_write_sample
    sample_offset = sampling_freq//target_clk

    cropStart = start_sample - 15
    cropEnd   = cropStart + 20
    print("t:", trigger_sim_ns+((cropStart+25)/200)*10, "\n",file=flog,flush=True)
    print(cropStart, cropEnd, "\n",file=flog,flush=True)

    # Splice traces
    spliced_traces = np.zeros([splice_depth*traces,cropEnd-cropStart])

    for s in range(splice_depth):
        traceSet = traceset.TraceSet(traceNum=traces,
                            fileName=TRACES_FILE,
                            cropStart = cropStart+s*sample_offset,
                            cropEnd = cropEnd+s*sample_offset)
        for i in range(traces):    
            spliced_traces[splice_depth*i+s,:] = traceSet.traces[i]

    traces_spliced = traces * splice_depth
    traceSet.traces = spliced_traces

    # ---------
    # Attack
    # ---------
    now = datetime.now()
    current_time = now.strftime("%H:%M:%S")
    print("Start Time =", current_time, "\n",file=flog,flush=True)

    measuredPower = traceSet.traces

    analysisDir = results_dir+"/attack_{}a".format(0)
    if not os.path.exists(analysisDir):
        os.makedirs(analysisDir)

    start_t = time.time()
    cpaAttacker = cpa.CPA()
    correctKey, mtd = cpaAttacker.doCPA(measuredPower=measuredPower,
                          hypotheticalPower=hypotheticalPower,
                          numTraces=traces_spliced,
                          analysisDir=analysisDir,
                          MTDStride= splice_depth*20,
                          numKeys = 1,
                          plot=True, 
                          plotSize=(10,8),
                          plotFontSize=18
                          )
    print(correctKey, mtd)
    mtd_intersections.append(mtd)
    correctKey = int(correctKey[0], 16)
    # Writing analysis configuration to file.
    end_t = time.time()
    print(end_t - start_t, "\n",file=flog,flush=True)

    now = datetime.now()
    current_time = now.strftime("%H:%M:%S")
    print("Finish Time =", current_time, "\n",file=flog,flush=True)

    abs_stop = time.time()
    print("Current Runtime Time =", abs_stop - abs_start, "\n",file=flog,flush=True)

    known_keys.append(correctKey)

    print("Keys: ", known_keys, [hex(x) for x in known_keys], "\n",file=flog,flush=True)
    print("Partials: ", known_partials, [hex(x) for x in known_partials], "\n",file=flog,flush=True)


    with open(results_dir+"/mtds.txt","w+") as f:
        f.write(f"[{splice_depth},{mtd}],")

    with open("keys.txt","w+") as f:
        for key in known_keys:
            f.write(f"{key},")

    h0_offset = h0_expected[0] - known_keys[0]

    if (h0_expected[0] != known_keys[0]):
        with open("failure.txt","w+") as f:
            for key in known_keys:
                f.write(f"{key},")
            f.write("\n\n")

            for key in h0_expected:
                f.write(f"{key},")
            f.write("\n\n")

            f.write("{} != {}". format(h0_expected[0],known_keys[0]))


    ##############################################
    # Attack first read to recover h0_msb[1]
    ##############################################
    # Attack configuration
    traces            = 5
    splice_depth      = math.ceil(r/32)-1

    # ---------
    # Power Model
    # ---------
    hypotheticalPower = [powermodel.getHypotheticalPower(PLAIN_TEXT, traces, splice_depth, 0, [0], [(known_keys[0]//32)*32], r)]

    # ---------
    # Get Traces
    # ---------
    ### Crop to target cycle
    print("Reading power traces...", "\n",file=flog,flush=True)
    start_sample  = target_read_sample
    sample_offset = sampling_freq//target_clk

    cropStart = start_sample - 35
    cropEnd   = cropStart + 60
    print("t:", trigger_sim_ns+((cropStart+25)/200)*10, "\n",file=flog,flush=True)
    print(cropStart, cropEnd, "\n",file=flog,flush=True)

    # Splice traces
    spliced_traces = np.zeros([splice_depth*traces,cropEnd-cropStart])

    for s in range(splice_depth):
        traceSet = traceset.TraceSet(traceNum=traces,
                            fileName=TRACES_FILE,
                            cropStart = cropStart+s*sample_offset,
                            cropEnd = cropEnd+s*sample_offset)
        for i in range(traces):    
            spliced_traces[splice_depth*i+s,:] = traceSet.traces[i]

    traces_spliced = traces * splice_depth
    traceSet.traces = spliced_traces

    # ---------
    # Attack
    # ---------
    now = datetime.now()
    current_time = now.strftime("%H:%M:%S")
    print("Start Time =", current_time, "\n",file=flog,flush=True)

    measuredPower = traceSet.traces

    analysisDir = results_dir+"/attack_{}b".format(0)
    if not os.path.exists(analysisDir):
        os.makedirs(analysisDir)

    start_t = time.time()
    cpaAttacker = cpa.CPA()
    correctKey, mtd = cpaAttacker.doCPA(measuredPower=measuredPower,
                          hypotheticalPower=hypotheticalPower,
                          numTraces=traces_spliced,
                          analysisDir=analysisDir,
                          MTDStride= 200,
                          numKeys = 1,
                          plot=True, 
                          plotSize=(10,8),
                          plotFontSize=18
                          )

    mtd_intersections.append(mtd)
    correctKey = int(correctKey[0], 16)
    # Writing analysis configuration to file.
    end_t = time.time()
    print(end_t - start_t, "\n",file=flog,flush=True)

    now = datetime.now()
    current_time = now.strftime("%H:%M:%S")
    print("Finish Time =", current_time, "\n",file=flog,flush=True)

    abs_stop = time.time()
    print("Current Runtime Time =", abs_stop - abs_start, "\n",file=flog,flush=True)

    known_partials.append((correctKey//32)*32)

    print("Keys: ", known_keys, [hex(x) for x in known_keys], "\n",file=flog,flush=True)
    print("Partials: ", known_partials, [hex(x) for x in known_partials], "\n",file=flog,flush=True)


    with open(results_dir+"/mtds.txt","a") as f:
        f.write(f"[{splice_depth},{mtd}],")

    with open("keys.txt","w+") as f:
        for key in known_keys:
            f.write(f"{key},")

    if (h0_expected[1]//32 != known_partials[0]//32):
        with open("failure.txt","w+") as f:
            for key in known_keys:
                f.write(f"{key},")
            f.write("\n\n")

            for key in h0_expected:
                f.write(f"{key},")
            f.write("\n\n")

            f.write("{} != {}". format(h0_expected[1]//32,known_partials[0]//32))


    #######################################################################
    ## Attack following round_i read to recover h0_msb[i] and h0_lsb[i-1]
    #######################################################################
    for attack_round in range(len(known_keys),w-1):
        retry = 1

        # Attack configuration
        traces            = 5
        splice_depth      = math.ceil(r/32)-1

        while (retry == 1):

            print("\n\n\n",file=flog,flush=True)
            print("-------------------------------------------",file=flog,flush=True)
            print("-------------------------------------------",file=flog,flush=True)
            print("-------------------------------------------",file=flog,flush=True)
            print("Round:", attack_round, "\n",file=flog,flush=True)
            print("Round:", attack_round, "\n")#file=flog,flush=True)
            mult_target_round = len(known_keys)

            ################################################
            ## Create hypothetical power model
            ################################################
            hypotheticalPower = [powermodel.getHypotheticalPower(PLAIN_TEXT, traces, splice_depth, mult_target_round, known_keys, known_partials, r)]

            ################################################
            ## Crop traces
            ################################################
            ############### No Reg
            round_offset_sample = round_len_sample*mult_target_round

            ### Crop to target cycle
            start_sample  = target_read_sample
            sample_offset = sampling_freq//target_clk

            cropStart = start_sample + round_offset_sample + 20
            cropEnd   = cropStart + 20

            print("t:", trigger_sim_ns+((cropStart+25)/200)*10, round_offset_sample/200*10, "\n",file=flog,flush=True)
            print(cropStart, cropEnd, "\n",file=flog,flush=True)

            # Splice traces
            spliced_traces = np.zeros([splice_depth*traces,cropEnd-cropStart])

            for s in range(splice_depth):
                traceSet = traceset.TraceSet(traceNum=traces,
                                    fileName=TRACES_FILE,
                                    cropStart = cropStart+s*sample_offset,
                                    cropEnd = cropEnd+s*sample_offset)
                for i in range(traces):    
                    spliced_traces[splice_depth*i+s,:] = traceSet.traces[i]

            traces_spliced = traces * splice_depth

            ################################################
            ## Run the attack
            ################################################
            measuredPower =     spliced_traces
            print(traces_spliced, "\n",file=flog,flush=True)
            print(len(measuredPower), "\n",file=flog,flush=True)
            print(len(hypotheticalPower[0]), "\n",file=flog,flush=True)

            now = datetime.now()
            current_time = now.strftime("%H:%M:%S")
            print("Start Time =", current_time, "\n",file=flog,flush=True)

            analysisDir = results_dir+"/attack_{}".format(mult_target_round)
            if not os.path.exists(analysisDir):
                os.makedirs(analysisDir)

            start_t = time.time()
            cpaAttacker = cpa.CPA()
            correctKey, mtd = cpaAttacker.doCPA(measuredPower=measuredPower,
                                  hypotheticalPower=hypotheticalPower,
                                  numTraces=traces_spliced,
                                  analysisDir=analysisDir,
                                  MTDStride= splice_depth*4,
                                  numKeys = 1,
                                  plot=True, 
                                  plotSize=(10,8),
                                  plotFontSize=18
                                  )
            
            mtd_intersections.append(mtd)

            correctKey = int(correctKey[0], 16)
            # Writing analysis configuration to file.
            end_t = time.time()
            print(end_t - start_t, "\n",file=flog,flush=True)

            now = datetime.now()
            current_time = now.strftime("%H:%M:%S")
            print("Finish Time =", current_time, "\n",file=flog,flush=True)
            
            abs_stop = time.time()
            print("Current Runtime Time =", abs_stop - abs_start, "\n",file=flog,flush=True)

            print("New Key: {}+{}={}".format(known_partials[-1],(correctKey%32),known_partials[-1]+(correctKey%32)), "\n",file=flog,flush=True)
            print("New Partial: {}// -> {}".format(correctKey, (correctKey//32)*32), "\n",file=flog,flush=True)

            known_keys.append(known_partials[-1]+(correctKey%32))
            known_partials.append((correctKey//32)*32)


            print("Expected Keys: ", h0_expected, "\n",flush=True)
            print("Keys: ", known_keys, "\n",flush=True)
            print("Keys (offset): ", [x+h0_offset for x in known_keys],  "\n",flush=True)
            print("Partials: ", known_partials,"\n",flush=True)
            print("Keys: ", known_keys, [hex(x) for x in known_keys], "\n",file=flog,flush=True)
            print("Partials: ", known_partials, [hex(x) for x in known_partials], "\n",file=flog,flush=True)

            with open(results_dir+"/mtds.txt","a") as f:
                f.write(f"[{splice_depth},{mtd}],")

            with open(results_dir+"/keys.txt","w+") as f:
                for key in known_keys:
                    f.write(f"{key},")

            print("{} != {}? {} != {}?".format(h0_expected[attack_round], known_keys[attack_round]+h0_offset, known_partials[-1]+(h0_offset//32)*32,(h0_expected[attack_round+1]//32)*32))
            if (h0_expected[attack_round] != (known_keys[attack_round]+h0_offset) or (known_partials[-1]+(h0_offset//32)*32) != (h0_expected[attack_round+1]//32)*32):
                with open("failure.txt","w+") as f:
                    for key in known_keys:
                        f.write(f"{key},")
                    f.write("\n\n")

                    for key in h0_expected:
                        f.write(f"{key},")
                    f.write("\n\n")

                    f.write("{} != {}". format(h0_expected[attack_round],known_keys[attack_round]))
                
                known_keys     = known_keys[:-1]
                known_partials = known_partials[:-1]
                mtd_intersections = mtd_intersections[:-1]

                traces += 10

                if (traces >= 100):
                    print("ATTACK FAILED")
                    return
            else:
                retry = 0
            
            print("-------------\n\n FINISHED ROUND--------------\n\n", flush=True)


    with open(results_dir+"/final_keys.txt","w+") as f:
        for key in known_keys:
            f.write(f"{key},")


    flog.close()
